// StepperLib.c : Defines the entry point for the DLL application.
//





#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <time.h>


#include <unistd.h>
#include <sys/sem.h>		//Used for semaphores
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>


#include "StepperLib.h"


#define TCHAR char

#define VSNPRINTF(a,b,c,d,e) ((((e)=vsnprintf((a),(b),(c),(d)))<0)?((a)[(b)-1]=0,(e)):(e))

void xcprintf(char * format, ...)
{
 va_list   args;
 char      v[512];
 int       numw;
 
 
 va_start(args, format);
 numw=VSNPRINTF(v,(sizeof(v)-2)/sizeof(TCHAR),format,args,numw);
 strcat(v,"\r\n");
 fprintf(stderr,"%s",v);
 va_end(args);
}




#ifndef _DEBUG


#define cprintf xcprintf

#else

extern int zcprintf(char * format, ...);


#define cprintf 1 || zcprintf

#endif









// http://www.raspberry-projects.com/pi/programming-in-c/semaphores/using-semaphores


union semun {
	int val;
	struct semid_ds *buf;
	unsigned short *array;
};


static int semaphore1_id;

//***********************************************************
//***********************************************************
//********** WAIT IF NECESSARY THEN LOCK SEMAPHORE **********
//***********************************************************
//***********************************************************
//Wait if necessary and then change the semaphore by 1. This is the "wait" operation
static int semaphore1_get_access(void)
{
	struct sembuf sem_b;

	sem_b.sem_num = 0;
	sem_b.sem_op = -1; /* P() */
	sem_b.sem_flg = SEM_UNDO | IPC_NOWAIT;
	if (semop(semaphore1_id, &sem_b, 1) == -1)		//Wait until free
	{
//		fprintf(stderr, "semaphore1_get_access failed\n");
		return(0);
	}
	return(1);
}

//***************************************
//***************************************
//********** RELEASE SEMAPHORE **********
//***************************************
//***************************************
//Setting the semaphore back to available.  This is the "release" operation.

static int semaphore1_release_access(void)
{
	struct sembuf sem_b;
	sem_b.sem_num = 0;
	sem_b.sem_op = 1; /* V() */
	sem_b.sem_flg = SEM_UNDO;
	if (semop(semaphore1_id, &sem_b, 1) == -1)
	{
		fprintf(stderr, "semaphore1_release_access failed\n");
		return(0);
	}
	return(1);
}




	//----------------------------
	//----- CREATE SEMAPHORE -----
	//----------------------------
void createsemaphore(void)
{
	cprintf("Creating semaphore...\n");
	semaphore1_id = semget((key_t)82312345, 1, 0666 | IPC_CREAT);		//<<<<< SET THE SEMPAHORE KEY   (Semaphore key, number of semaphores required, flags)
	//	Semaphore key
	//		Unique non zero integer (usually 32 bit).  Needs to avoid clashing with another other processes semaphores (you just have to pick a random value and hope - ftok() can help with this but it still doesn't guarantee to avoid colision)

	//Initialize the semaphore using the SETVAL command in a semctl call (required before it can be used)
	union semun sem_union_init;
	sem_union_init.val = 1;
	if (semctl(semaphore1_id, 0, SETVAL, sem_union_init) == -1)
	{
		fprintf(stderr, "Creating semaphore failed to initialize\n");
		exit(EXIT_FAILURE);
	}
}


void deletesemaphore(void)
{

	//----------------------------
	//----- DELETE SEMAPHORE -----
	//----------------------------
	//It's important not to unintentionally leave semaphores existing after program execution. It also may cause problems next time you run the program.
	union semun sem_union_delete;
	if (semctl(semaphore1_id, 0, IPC_RMID, sem_union_delete) == -1)
		fprintf(stderr, "Failed to delete semaphore\n");

}




#define MILLIS (CLOCKS_PER_SEC/1000)

#define DEFDELAY (5000*MILLIS)



static int resetflag=0;

static int stepperactive=0;



#define OUR_INPUT_FIFO_NAME "/tmp/stepperout"
#define OUR_OUTPUT_FIFO_NAME "/tmp/stepperin"

int our_input_fifo_filestream = -1;
int our_output_fifo_filestream = -1;


int ipread(char * p,int n)
{
 int r;

 
 r=read(our_input_fifo_filestream,p,n);

// if(r>=0) cprintf("ipread n %d r %d",n,r);
// else
 if(r<0 && errno!=EAGAIN) cprintf("error %d\n",errno);
 
 return(r);
}    
    

int ipwrite(char * p,int n)
{
 int w;
 
 w=write(our_output_fifo_filestream,p,n);
    
 return(w);   
}
    
    
    
static char * readlnbase(int delay)
{
 static char buffer[1024];
 static char * p=buffer;
 static char * q=buffer;
 int           n;
 clock_t       timeout;
 


 timeout=delay*MILLIS+clock();

 while(1)
 {
  if(resetflag) return(NULL);

  n=ipread(p,1);

  if(n<=0)
  {
   if(clock()>timeout) return(NULL);
  }
  else
  {
//   cprintf("n chars %d ",*p);

   p+=n;
   *p=0;

//   cprintf("buffer %s",buffer);

   if((p-buffer+4)>=sizeof(buffer))
   {
    p=q=buffer;
    return(buffer);
   }

   while(q<p)
   {
    if(*q++<32)
    {
//     cprintf("read ln term %d",*(q-1));
     *(q-1)=0;
      p=q=buffer;
     if(*buffer)
     {
      return(buffer);
     }
    }
   }
  }
 }
}



#define MAXMOTORS 3


static int seenmove[MAXMOTORS];
static int moveposn[MAXMOTORS];

static char * cachedcommand=NULL;
static char   cachebuffer[1024];



static char * readlnmiddle(int delay)
{
 char * p;
 int    posn;
 int    power;
	int    motor;
 int    cn;


 p=readlnbase(delay);
 if(p)
 {
//  cprintf("rlm %s",p);
  if((cn=sscanf(p,"mv %d %d %d",&motor,&posn,&power))==3)
  {
   if(motor>=0 && motor<MAXMOTORS)
   {
    seenmove[motor]=1;
    moveposn[motor]=posn;
   }
  }
  else
  {
   strcpy(cachebuffer,p);
   cachedcommand=cachebuffer;
  }
//  cprintf("cn %d seenmove %d",cn,seenmove);
 }

 return(p);
}




int TestMoveOver(int motor,int * posn,int delay)
{
 clock_t timeout=delay*MILLIS+clock();


// cprintf("tm in %d",seenmove);

 while(clock()<timeout)
 {
  if(seenmove[motor]) break;
  if(resetflag) break;

  if(semaphore1_get_access())
  {
   if(!seenmove[motor]) readlnmiddle(delay);
   semaphore1_release_access();
  }
  usleep(100*1000);
 }

// cprintf("tm ex %d",seenmove);

 if(resetflag) return(2);
 if(posn) *posn=moveposn[motor];
 return(seenmove[motor]);
}







char * readln(int delay)
{
 char * p;
 clock_t timeout=delay*MILLIS+clock();


 while(clock()<timeout)
 {
  if(cachedcommand) break;
  if(resetflag) break;

cprintf("read ln 1");

  if(semaphore1_get_access())
  {
	  cprintf("read ln 3");
   if(!cachedcommand) readlnmiddle(delay);
   semaphore1_release_access();
  }
  usleep(100*1000);
 }

 cprintf("read ln 2");

 if(resetflag) return(NULL);
 p=cachedcommand;
 cachedcommand=NULL;
 return(p); 
}




int writeserialport(int delay,char * format, ...)
{
 char v[1024];
 int  numw;
 va_list   args;

 if(resetflag) return(2);

 if(semaphore1_get_access())
 {
  va_start(args, format);
  numw=VSNPRINTF(v,(sizeof(v)-2)/sizeof(TCHAR),format,args,numw);

  va_end(args);

  ipwrite(v,strlen(v));

  semaphore1_release_access();
 }

 return(0);
}




STEPPER_API int ResetStepper(unsigned int timeout)
{
 cprintf("doing reset");
 
 if(semaphore1_get_access())
 {
  semaphore1_release_access();
 }

 return(0);
}



STEPPER_API int SetCancelState(int state)
{
 cprintf("set cancel flag %d",state);

 resetflag=state;

 return(0);
}





char * writeresponse(int delay,char * format, ...)
{
 char v[1024];
 int  numw;
 va_list   args;
 clock_t timeout=delay*MILLIS+clock();
 char * p=NULL;


 if(resetflag) return(NULL);

 if(semaphore1_get_access())
 {
  va_start(args, format);
  numw=VSNPRINTF(v,(sizeof(v)-2)/sizeof(TCHAR),format,args,numw);

  va_end(args);

  if(stepperactive)
  {
   cprintf("write response");

   ipwrite(v,strlen(v));


   while(clock()<timeout)
   {
    if(resetflag) break;
    if(cachedcommand) break;
    readlnmiddle(delay);
   }
   p=cachedcommand;
   cachedcommand=NULL;
  }
  semaphore1_release_access();
 }



 return(p);
}



int dbprint(char * string)
{
 cprintf(string);
 return(0);
}





STEPPER_API int RunSpeed(int motor,int speed)
{
 cprintf("run speed %d %d",motor,speed);

 writeserialport(DEFDELAY,"rs %d %d\n",motor,speed);

 return(0);
}






STEPPER_API int SetSpeedMotor(int motor,int speed)
{
 cprintf("speed motor %d %d",motor,speed);

 writeserialport(DEFDELAY,"sp %d %d\n",motor,speed);

 return(0);
}




STEPPER_API int SetAccelerationMotor(int motor,int acc)
{
 cprintf("acc motor %d %d",motor,acc);

 writeserialport(DEFDELAY,"ac %d %d\n",motor,acc);

 return(0);
}




STEPPER_API int MoveMotor(int motor,int steps)
{
 cprintf("move motor %d %d",motor,steps);
 seenmove[motor]=0;

 writeserialport(DEFDELAY,"mv %d %d\n",motor,steps);

 return(0);
}




STEPPER_API int MoveToMotor(int motor,int steps)
{
 cprintf("move To motor 1 %d",steps);
 seenmove[motor]=0;

 writeserialport(DEFDELAY,"mt %d %d\n",motor,steps);

 return(0);
}


STEPPER_API int StopMotor(int motor)
{
 cprintf("stop motor %d",motor);

 writeserialport(DEFDELAY,"xx %d\n",motor);

 return(0);
}



STEPPER_API int SetCurPos(int motor,int steps)
{
 cprintf("set cur pos %d %d",motor,steps);

 writeserialport(DEFDELAY,"sc %d %d\n",motor,steps);

 return(0);
}



STEPPER_API int StepSizeMotor(int motor,int steps)
{
 cprintf("step size motor %d %d",motor,steps);

 writeserialport(DEFDELAY,"ss %d %d\n",motor,steps);

 return(0);
}



STEPPER_API int PowerMotor(int motor,int pw)
{
 cprintf("power motor %d %d",motor,pw);

 writeserialport(DEFDELAY,"pw %d\n",pw);

 return(0);
}



STEPPER_API int WritePin(int pin,int value)
{
 cprintf("write pin %d %d",pin,value);

 writeserialport(DEFDELAY,"wr %d %d\n",pin,value);

 return(0);
}


STEPPER_API int ArmInterrupt(int motor,int interrupt,int mode)
{
 cprintf("arm interrupt %d %d %d",motor,interrupt,mode);

 writeserialport(DEFDELAY,"ai %d %d %d\n",motor,interrupt,mode);

 return(0);
}



STEPPER_API int RemInterrupt(int motor,int interrupt)
{
 cprintf("rem interrupt %d %d",motor,interrupt);

 writeserialport(DEFDELAY,"ri %d %d\n",motor,interrupt);

 return(0);
}



STEPPER_API int SendCommand(char * string)
{
 cprintf("send command %s",string);

 writeserialport(DEFDELAY,"%s\n",string); 

 return(0);
}



STEPPER_API int GetResponse(char ** string,int delay)
{
 char * p;

 p=readln(delay);
 
 if(p)
 {
  cprintf("get response %d %s len=%d",delay,p,strlen(p));
  *string=p;
  return(0);
 }

 return(1);
}




STEPPER_API int GetPosition(int motor)
{
 char * p;
 int    value;

 cprintf("get position function");
 if(!stepperactive) return(0);

 p=writeresponse(DEFDELAY,"ps %d\n",motor);
 if(p)
 {
  value=0;
  cprintf("get pos result %s",p);
  sscanf(p,"ps %d",&value);
  return(value);
 }

 cprintf("get pos times out");
 
 return(0);
}



STEPPER_API int GetDistanceToGo(int motor)
{
 char * p;
 int    value;


 cprintf("get distance to go %d",motor);
 p=writeresponse(DEFDELAY,"ds %d\n",motor);
 if(p)
 {
  value=0;
  sscanf(p,"ds %d",&value);
  return(value);
 }
 return(0);
}



STEPPER_API int ReadPin(int pin)
{
 char * p;
 int    value;

 cprintf("read pin %d",pin);

 p=writeresponse(DEFDELAY,"rd %d\n",pin);
 if(p)
 {
  value=0;
  sscanf(p,"rd %d",&value);
  cprintf("read pin %d value %d string %s",pin,value,p);
  return(value);
 }
 return(0);
}




STEPPER_API int TestSensor(int pin)
{
 char * p;
 int    value;


 cprintf("test pin %d",pin);

 p=writeresponse(DEFDELAY,"ts %d\n",pin);
 if(p)
 {
  value=0;
  sscanf(p,"ts %d",&value);
  cprintf("test pin %d value %d string %s",pin,value,p);
  return(value);
 }
 return(0);
}



STEPPER_API int FinitStp()
{
 deletesemaphore(); 
 return(0);
}



STEPPER_API int InitStp()
{
 createsemaphore();

 our_input_fifo_filestream=open(OUR_INPUT_FIFO_NAME, (O_RDONLY | O_NONBLOCK));
 if (our_input_fifo_filestream!=-1) cprintf("Opened FIFO: %i\n", our_input_fifo_filestream);
 else                               cprintf("error %d\n",errno);

 our_output_fifo_filestream=open(OUR_OUTPUT_FIFO_NAME, (O_WRONLY  | O_NONBLOCK));
 if(our_output_fifo_filestream!=-1) cprintf("Opened FIFO: %i\n", our_output_fifo_filestream);
 else                               cprintf("error %d\n",errno); 

 if(our_input_fifo_filestream!=-1 && our_output_fifo_filestream!=-1) stepperactive=1;
 
 return(0);
}

